{-# LANGUAGE EmptyDataDecls, FlexibleInstances, TypeOperators #-}
module AST where

import Control.Applicative((<$>))
import Data.Function
import Data.Map(Map)
import qualified Data.Map as M
import Data.Monoid
import Data.Set(Set)
import qualified Data.Set as S
import Data.Ord

import Engine
import Instr
import Label
import Type

infix  2 :::
infixl 4 :$
infixl 3 :@

type Var  = String
type Node = String
type Name = String

data Associativity = L | R
  deriving (Eq, Ord, Show)
type Precedence    = Int
data Fixity
  = Infix Associativity Precedence
  | Prefix
  deriving (Eq, Ord, Show)

data Constant = Constant
  { cName   :: Name
  , cFixity :: Fixity
  , cImpl   :: Either (Type -> LabelGen Engine
                      ,Int -- Number of type variables needed
                      ,[Type] -> Type)
                      Term
  }
instance Eq Constant where
  (==) = (==) `on` cName
instance Show Constant where
  show = cName
instance Ord Constant where
  compare = comparing cName

data BaseTerm term
  = Var Var
  | Abs Var (Maybe Type) term
  | term :$ term
  | Lit Lit
  | Const Constant
  | Pair
  | Proj1
  | Proj2
  | Fix
  | term :@ Node
  deriving (Eq, Ord, Show)

instance TypeSubst term => TypeSubst (BaseTerm term) where
  typeSubst sub term = case term of
    Var v      -> Var v
    Abs v mt t -> Abs v (rec <$> mt) $ rec t
    t1 :$ t2   -> rec t1 :$ rec t2
    Lit n      -> Lit n
    Const c    -> Const c
    Pair       -> Pair
    Proj1      -> Proj1
    Proj2      -> Proj2
    Fix        -> Fix
    t :@ n     -> rec t :@ n
    -- Case t b d -> Case (rec t) (map (second rec) b) (rec d)
    where
      rec :: TypeSubst a => a -> a
      rec = typeSubst sub

data a ::: b = a ::: b
  deriving (Eq, Ord, Show)
instance (TypeSubst a, TypeSubst b) => TypeSubst (a ::: b) where
  typeSubst sub (a ::: b) = typeSubst sub a ::: typeSubst sub b

newtype Term = Term (BaseTerm Term)
  deriving (Eq, Ord, Show)

unTerm :: Term -> BaseTerm Term
unTerm (Term t) = t
instance TypeSubst Term where
  typeSubst sub (Term t) = Term $ typeSubst sub t

type Substitution a = Map Var a

class Subst a where
  subst :: Substitution a -> a -> a
class FreeVars a where
  fv :: a -> Set Var

instance Subst Term where
  subst sub (Term term) = Term $ case term of
    Var v      -> unTerm $ M.findWithDefault (Term $ Var v) v sub
    Abs v mt t | v `S.notMember` sup -> Abs v mt $ subst (M.delete v sub) t
               | otherwise           -> Abs v' mt $ subst (M.insert v (Term $ Var v') sub) t
      where
        sup = S.unions $ M.elems $ M.map fv sub
        v' = head $ dropWhile (`S.member` sup) [show n ++ v | n <- [(0 :: Integer)..]]
    t1 :$ t2   -> rec t1 :$ rec t2
    Lit n      -> Lit n
    Const c    -> Const c
    Pair       -> Pair
    Proj1      -> Proj1
    Proj2      -> Proj2
    Fix        -> Fix
    t :@ n     -> rec t :@ n
    where
      rec :: Term -> Term
      rec = subst sub

instance FreeVars term => FreeVars (BaseTerm term) where
  fv term = case term of
    Var v     -> S.singleton v
    Abs v _ t -> S.delete v $ fv t
    t1 :$ t2  -> fv t1 `mappend` fv t2
    Lit _     -> mempty
    Const _   -> mempty
    Pair      -> mempty
    Proj1     -> mempty
    Proj2     -> mempty
    Fix       -> mempty
    t :@ _    -> fv t

instance FreeVars TypedTerm where
  fv (TypedTerm t) = fv t

instance FreeVars t => FreeVars (t ::: a) where
  fv (x ::: _) = fv x

instance FreeVars Term where
  fv (Term t) = fv t

newtype TypedTerm = TypedTerm (BaseTerm (TypedTerm ::: Type))
  deriving (Eq, Ord, Show)
unTypedTerm :: TypedTerm -> BaseTerm (TypedTerm ::: Type)
unTypedTerm (TypedTerm t) = t
instance TypeSubst TypedTerm where
  typeSubst sub (TypedTerm t) = TypedTerm $ typeSubst sub t
