module Constants where

import Data.Map(Map)
import qualified Data.Map as M

import AST
import Engine
import Instr
import Interface
import Pretty
import Type

constants :: Map Name Constant
constants = M.fromList $ zip (map cName cs) cs
  where
    cs =
      -- Numeric operations
      [binop "BINOP" "+"  (Infix L 6) "+"  1 (\[a] -> a :=>: a :=>: a)
      ,binop "BINOP" "-"  (Infix L 6) "-"  1 (\[a] -> a :=>: a :=>: a)
      ,binop "BINOP" "*"  (Infix L 7) "*"  1 (\[a] -> a :=>: a :=>: a)
      ,binop "BINOP" "/"  (Infix L 7) "/"  1 (\[a] -> a :=>: a :=>: a)
      ,binop "BINOP_ALL" "==" (Infix L 3) "==" 1 (\[a] -> a :=>: a :=>: tint)
      ,binop "BINOP_ANY" "!=" (Infix L 3) "!=" 1 (\[a] -> a :=>: a :=>: tint)
      ,binop "BINOP_ALL" "<"  (Infix L 3) "<"  1 (\[a] -> a :=>: a :=>: tint)
      ,binop "BINOP_ALL" "<=" (Infix L 3) "<=" 1 (\[a] -> a :=>: a :=>: tint)
      ,binop "BINOP_ALL" ">"  (Infix L 3) ">"  1 (\[a] -> a :=>: a :=>: tint)
      ,binop "BINOP_ALL" ">=" (Infix L 3) ">=" 1 (\[a] -> a :=>: a :=>: tint)
      ,binop "BINOP" "&"  (Infix L 5) "&&" 0 (\[]  -> tint :=>: tint :=>: tint)
      ,binop "BINOP" "|"  (Infix L 4) "||" 0 (\[]  -> tint :=>: tint :=>: tint)

      -- Some math functions
      ,floatfun "cos" "cos"
      ,floatfun "sin" "sin"
      ,floatfun "tan" "tan"
      ,floatfun "acos" "acos"
      ,floatfun "asin" "asin"
      ,floatfun "atan" "atan"
      ,floatfun "cosh" "cosh"
      ,floatfun "sinh" "sinh"
      ,floatfun "tanh" "tanh"
      ,floatfun "exp" "exp"
      ,floatfun "log" "log"
      ,floatfun "log10" "log10"
      ,floatfun "sqrt" "sqrt"
      ,floatfun "ceil" "ceil"
      ,floatfun "fabs" "fabs"
      ,floatfun "floor" "floor"
      ,floatbinfun "atan2" "atan2"
      ,floatbinfun "pow" "pow"
      ,floatbinfun "fmod" "fmod"
      -- Casting
      ,floatintfun "int" "(int_t)"
      ,intfloatfun "float" "(floatptr_t)"
      ,intfuncom "print_int" "print_int"
      ,floatfuncom "print_float" "print_float"
      -- Pairs
      ,Constant
        { cName   = ","
        , cFixity = Infix R 1
        , cImpl   = Right $ Term $ Pair
        }
      ,Constant
        { cName   = "fst"
        , cFixity = Prefix
        , cImpl   = Right $ Term $ Proj1
        }
      ,Constant
        { cName   = "snd"
        , cFixity = Prefix
        , cImpl   = Right $ Term $ Proj2
        }
      -- Commands
      ,Constant
        { cName   = ";"
        , cFixity = Infix R 0
        , cImpl   = Left (const seqEngine
                         , 1, \[a] -> tcom :=>: a :=>: a)
        }
      ,Constant
        { cName   = "skip"
        , cFixity = Prefix
        , cImpl   = Left (const skipEngine
                         , 0, \[] -> tcom)
        }
      -- Fix
      ,Constant
        { cName   = "fix"
        , cFixity = Prefix
        , cImpl   = Right $ Term $ Fix
        }
      -- Variables
      ,Constant
        { cName   = "new"
        , cFixity = Prefix
        , cImpl   = Left (const newvarEngine
                         , 2, \[a, b] -> (tvar a :=>: b) :=>: b)
        }
      ,Constant
        { cName   = "newarray"
        , cFixity = Prefix
        , cImpl   = Left (const newarrayEngine
                         , 2, \[a, b] -> tint :=>: (tmutarray a :=>: b) :=>: b)
        }
      ,Constant
        { cName   = "!"
        , cFixity = Prefix
        , cImpl   = Right $ Term Proj1
        }
      ,Constant
        { cName   = ":="
        , cFixity = Infix R 2
        , cImpl   = Right $ Term Proj2
        }
      -- If then else
      ,Constant
        { cName   = "if then else"
        , cFixity = Prefix
        , cImpl   = Left (const ifEngine
                         , 1, \[a] -> tint :=>: a :=>: a :=>: a)
        }
      -- Par
      ,Constant
        { cName   = "||"
        , cFixity = Infix R 1
        , cImpl   = Left (const parEngine
                         , 0, \_ -> tcom :=>: tcom :=>: tcom)
        }
      -- Constant array operations
      ,Constant
        { cName = "singleton"
        , cFixity = Prefix
        , cImpl = Left (\(t :=>: _) -> idEngine (forgetVars t)
                       , 1, \[a] -> a :=>: tarray a)
        }
      {- ,Constant
        { cName = "length"
        , cFixity = Prefix
        , cImpl = Left (const lengthEngine
                       , 1, \[a] -> tarray a :=>: tint)
      ,Constant
        { cName = "index"
        , cFixity = Prefix
        , cImpl = Left (const indexEngine
                       , 1, \[a] -> tint :=>: tarray a :=>: a)
        }
      ,Constant
        { cName = "range"
        , cFixity = Prefix
        , cImpl = Left (const rangeEngine
                        , 1, \[a] -> tint :=>: tint :=>: tarray a :=>: tarray a)
        }
      ,Constant
        { cName = "++"
        , cFixity = Infix L 6
        , cImpl = Left (const appendEngine
                       , 1, \[a] -> tarray a :=>: tarray a :=>: tarray a
        }
      -}
      ]
      where
        field t = case t of
          TConstr TInt   []                  -> ("integer", "1")
          TConstr TFloat []                  -> ("floating", "1")
          TConstr TArray [TConstr TInt []]   -> ("integer", "regs->dataLength")
          TConstr TArray [TConstr TFloat []] -> ("floating", "regs->dataLength")
          _                 -> error $ "field: " ++ pretty t
        binop macro name fixity op tvars typfun = Constant
          { cName   = name
          , cFixity = fixity
          , cImpl   =
            Left (\(t1 :=>: t2 :=>: t3) ->
              let (f, len)  = field t1
                  opf x y d = macro ++ "(" ++ op ++ ", " ++ len ++ ", " ++ f ++ ")"
               in binOpEngine op (PrimOp opf), tvars, typfun)
          }
        floatfun name f = Constant
          { cName = name
          , cFixity = Prefix
          , cImpl = let opf x _ d = reg0 d ++ ".floating = "
                                    ++ f ++ "(" ++ reg0 x ++ ".floating)"
                     in Left (const (unOpEngine f (PrimOp opf))
                         , 0, const $ tfloat :=>: tfloat)
          }
        floatfuncom name f = Constant
          { cName = name
          , cFixity = Prefix
          , cImpl = let opf x _ d = f ++ "(" ++ reg0 x ++ ".floating)"
                     in Left (const (unOpEngine f (PrimOp opf))
                         , 0, const $ tfloat :=>: tcom)
          }
        intfuncom name f = Constant
          { cName = name
          , cFixity = Prefix
          , cImpl = let opf x _ d = f ++ "(" ++ reg0 x ++ ".integer)"
                     in Left (const (unOpEngine f (PrimOp opf))
                         , 0, const $ tint :=>: tcom)
          }
        floatbinfun name f = Constant
          { cName   = name
          , cFixity = Prefix
          , cImpl   = let opf x y d = reg0 d ++ ".floating = "
                                   ++ f ++ "(" ++ reg x ++ ".floating, "
                                   ++ reg0 y ++ ".floating)"
                       in Left (const (binOpEngine f (PrimOp opf))
                           , 0, const $ tfloat :=>: tfloat :=>: tfloat)
          }
        floatintfun name f = Constant
          { cName = name
          , cFixity = Prefix
          , cImpl = let opf x _ d = reg0 d ++ ".integer = "
                                    ++ f ++ "(" ++ reg0 x ++ ".floating)"
                     in Left (const (unOpEngine f (PrimOp opf))
                         , 0, const $ tfloat :=>: tint)
          }
        intfloatfun name f = Constant
          { cName = name
          , cFixity = Prefix
          , cImpl = let opf x _ d = reg0 d ++ ".floating = "
                                    ++ f ++ "(" ++ reg0 x ++ ".integer)"
                     in Left (const (unOpEngine f (PrimOp opf))
                         , 0, const $ tint :=>: tfloat)
          }

typeOperations :: [(Name, Type -> Type -> Type, Fixity)]
typeOperations =
  [("->", (:=>:), Infix R 0)
  ,("*", (:*:), Infix R 1)
  ]
