module Dot where
import Data.Function
import Data.List
import Data.Map(Map)
import qualified Data.Map as M
import Data.Maybe
import qualified Data.Set as S

import AST
import Engine
import Interface
import Label

infixr 5 <+>
infix 4 <$>
infix 6 =:

(<+>), (<$>) :: String -> String -> String
"" <+> t = t
s <+> "" = s
s <+> t = s ++ " " ++ t
s <$> t = s ++ "\n" ++ t

indent :: String -> String
indent = unlines . map ("  " ++) . lines

braces_, braces, brackets :: String -> String
braces_ s = "{" ++ s ++ "}"
braces s = braces_ $ "" <$> indent s
brackets s = "[" ++ s ++ "]"

digraph :: [String] -> String
digraph s = "digraph g" <+> braces (unlines $
  ["mc=2"
  ,"node [shape = plaintext];"
  ,"edge [arrowtail = none];"
  ] ++ s)

tag :: String -> String -> String -> String
tag x y s = "<" ++ x <+> y ++ ">" ++ s ++ "</" ++ x ++ ">"
table, td :: String -> String -> String
table x s = tag "TABLE" x s
td x s = tag "TD" x s
tr :: String -> String
tr ""   = ""
tr s    = tag "TR" "" s
(=:) :: String -> String -> String
x =: y = x ++ "=\"" ++ y ++ "\""
fontSize :: Int -> String -> String
fontSize size s = tag "FONT" ("POINT-SIZE" =: show size) s

pad :: Int -> a -> [a] -> [a]
pad len el xs = replicate (h `div` 2) el ++ xs ++ replicate ((h + 1) `div` 2) el
  where h = len - length xs

type DotNode = Int
type DotPort = Int

generateDot :: [(DotNode, String, [Label], [Label])]
            -> [((DotNode, Label), (DotNode, Label))]
            -> [String]
generateDot nodes connections = map node nodes ++ map connection connections
  where
    node (n, name, inPorts, outPorts) = structName <+> brackets ("label = <" ++ table tableAttribs
      (concat $ map tr $ map concat $ transpose $ map (pad height (td "" ""))
        [labelList inPorts, [td "" $ fontSize 12 $ name], labelList outPorts])
      ++ ">") ++ ";"
      where
        height = (max `on` length) inPorts outPorts
        structName = "struct" ++ show n
    tableAttribs = "BORDER" =: "1" <+> "CELLBORDER" =: "0" <+> "CELLSPACING" =: "0"
    connection ((n1, l1), (n2, l2))
      | n1 == n2  = ""
      | otherwise = "struct" ++ show n1 ++ ":" ++ show l1  <+> "->" <+>
                    "struct" ++ show n2 ++ ":" ++ show l2 ++ ";"
    labelList ls = [td ("PORT" =: sl) $ fontSize 7 sl | l <- ls, let sl = show l]


dotEngines :: Map Label Node -> Map Node Int -> [(DotNode, Engine)] -> String
dotEngines labelMappings nodeMappings es = digraph $ generateDot nlabels conn
  where
    nlabels = map nodeLabels es
    conn   =  [((node, olabel), (nodeNum', olabel))
              | (node, _, _ilabel, olabels) <- nlabels, olabel <- olabels
              , node'    <- catMaybes [M.lookup olabel labelMappings]
              , nodeNum' <- catMaybes [M.lookup node' nodeMappings]]
    nodeLabels (node, Engine name _loc _ls interface) =
      (node, show node ++ ":" ++ escape (limit name), opp, pla)
      where
        opp = S.toList $ opponent interface S.\\ player interface
        pla = S.toList $ player interface S.\\ opponent interface
        -- sameNode l = (flip M.lookup nodeMappings =<< M.lookup l labelMappings) == Just node
        escape = filter $ flip notElem ['<', '>']
        limit name | length name > 50 = take 47 name ++ "..."
                   | otherwise        = name
