module Instr where

-- import Interface
import Label

infixr 6 :.
infix 7 <~

data PrimLit
  = LInt Integer
  | LFloat Integer Integer
  deriving (Eq, Ord, Show)

data Lit
  = PrimLit PrimLit
  | LArray [PrimLit]
  deriving (Eq, Ord, Show)

litArrayView :: Lit -> [PrimLit]
litArrayView (PrimLit p) = [p]
litArrayView (LArray ls) = ls

lint :: Integer -> Lit
lint = PrimLit . LInt

lfloat :: Integer -> Integer -> Lit
lfloat x = PrimLit . LFloat x

data Reg = Register String | Index Reg String
         | Reference Reg   | Field Reg String
         | Dereference Reg
  deriving (Eq, Ord, Show)

reg :: Reg -> String
reg (Register r) = r
reg (Index r i)  = reg r ++ "[" ++ i ++ "]"
reg (Reference r) = "(&" ++ reg r ++ ")"
reg (Dereference r) = "(*" ++ reg r ++ ")"
reg (Field r f) = reg r ++ "." ++ f

reg0 :: Reg -> String
reg0 = reg . (`Index` "0")

p1Reg,  p2Reg, dataReg, dataReg0, dataLengthReg, emptyReg, tempReg, const0, const1 :: Reg
p1Reg = Register "regs->message.ptr1"
p2Reg = Register "regs->message.ptr2"
dataReg = Register "regs->data"
dataReg0 = dataReg `Index` "0"
dataLengthReg = Register "regs->dataLength"
tempReg = Register "regs->temp"
emptyReg = Register "emptyReg"

const0 = constReg 0
const1 = constReg 1

constReg :: (Num a, Show a) => a -> Reg
constReg n = Register $ show n

newtype PrimOp = PrimOp (Reg -> Reg -> Reg -> String)

-- Some dummy instances to be able to derive stuff for Instr
instance Eq   PrimOp where _ == _ = True
instance Ord  PrimOp where _ <= _ = True
instance Show PrimOp where show _ = "primOp"

data Instr
  = New Reg Reg Reg Reg
  | Get Reg Reg (Reg, Reg)
  | Update Reg Reg Reg
  | Free Reg
  | Flip Reg Reg
  | Set PrimLit Reg
  | Prim PrimOp Reg Reg Reg
  | Par Reg
  | Resize Reg
  | Spark Label
  deriving (Eq, Ord, Show)

instance LabelSubst Instr where
  labelSubst subst instr = case instr of
    New _ _ _ _  -> instr
    Get _ _ _    -> instr
    Update _ _ _ -> instr
    Free _       -> instr
    Flip _ _     -> instr
    Set _ _      -> instr
    Prim _ _ _ _ -> instr
    Par _        -> instr
    Resize _     -> instr
    Spark l      -> Spark $ labelSubst subst l

data Code
  = Instr :. Code
  | IfZero Reg Code Code
  | Unpar Reg Code
  | End
  deriving (Eq, Ord, Show)

instance LabelSubst Code where
  labelSubst subst code = case code of
    i :. c         -> rec i :. rec c
    IfZero r c1 c2 -> IfZero r (rec c1) (rec c2)
    Unpar r c      -> Unpar r (rec c)
    End            -> End
    where
      rec :: LabelSubst a => a -> a
      rec = labelSubst subst

sparkE :: Label -> Code
sparkE l = Spark l :. End

(<~) :: a -> (a -> b) -> b
i <~ k = k i

cci, ccq, cca, exi, exq :: Code -> Code
cci c = Flip p1Reg p2Reg :.  p2Reg <~ New p1Reg tempReg const1 :. c
ccq c = (p1Reg, tempReg) <~ Get p1Reg const1 :. p2Reg <~ New p2Reg tempReg const1 :. c
ccqe c = (p1Reg, emptyReg) <~ Get p1Reg const1 :. p2Reg <~ New p2Reg tempReg const1 :. c
cca c = Flip p1Reg p2Reg :. (p1Reg, tempReg) <~ Get p2Reg const1 :. Free p2Reg :. c
exi c = (p1Reg, tempReg) <~ Get p1Reg const1 :. p2Reg <~ New p2Reg p1Reg const1 :. c
exq c = (emptyReg, p1Reg) <~ Get p1Reg const1 :. p2Reg <~ New p2Reg tempReg const1 :. c
