{-# LANGUAGE BangPatterns, DeriveDataTypeable, TemplateHaskell #-}
module Lexer where

import Data.Char
import Data.Data
import Language.Haskell.TH.Lift
import Data.Parser.Grempa.Static

-- | Token datatype
data Tok
    = Var String
    | Con String
    | Op  String
    | Bang
    | If | Then | Else
    | Let | Equals | In
    | Fix
    | MapReduce
    | Macro
    | Num Integer
    | Dot
    | Colon
    | LParen | RParen
    | LBracket | RBracket
    | Lambda
    | At
  deriving (Eq, Ord, Data, Typeable, Show, Read)

present :: Tok -> String
present tok = case tok of
  Var _     -> "identifier"
  Con _     -> "type constructor"
  Op  _     -> "operator"
  Bang      -> "keyword '!'"
  If        -> "keyword 'if'"
  Then      -> "keyword 'then'"
  Else      -> "keyword 'else'"
  Let       -> "keyword 'let'"
  Equals    -> "keyword '='"
  In        -> "keyword 'in'"
  Fix       -> "keyword 'fix'"
  MapReduce -> "macro 'mapreduce'"
  Macro     -> "keyword 'macro'"
  Num _     -> "integer literal"
  Dot       -> "keyword '.'"
  Colon     -> "type annotation operator ':'"
  LParen    -> "left parenthesis '('"
  RParen    -> "right parenthesis ')'"
  LBracket  -> "left square bracket '['"
  RBracket  -> "right square bracket ']'"
  Lambda    -> "lambda operator '\\'"
  At        -> "locus specifier operator '@'"

fromTok :: Tok -> String
fromTok tok = case tok of
  Var s -> s
  Con s -> s
  Op  s -> s
  _     -> error $ "fromTok: " ++ show tok

fromNum :: Tok -> Integer
fromNum tok = case tok of
  Num n -> n
  _     -> error $ "fromNum: " ++ show tok

$(deriveLift ''Tok)
instance ToPat Tok where toPat = toConstrPat

-- * Shorthands for constructors applied to something
--   (could be anything since the ToPat instance creates wildcard patterns for
--    everything save for the constructor)
variable, constructor, operation, number :: Tok
variable    = Var ""
constructor = Con ""
operation   = Op  ""
number      = Num 0

-- | Do the lexing!
lexToks :: Int -> String -> [(Tok, Int)]
lexToks !n str = case str of
  [] -> []
  ('i':'f'        :as) | testHead (not . isId)  as -> If     ~> next 2 as
  ('t':'h':'e':'n':as) | testHead (not . isId)  as -> Then   ~> next 4 as
  ('e':'l':'s':'e':as) | testHead (not . isId)  as -> Else   ~> next 4 as
  ('l':'e':'t'    :as) | testHead (not . isId)  as -> Let    ~> next 3 as
  ('f':'i':'x'    :as) | testHead (not . isId)  as -> Fix    ~> next 3 as
  ('m':'a':'c':'r':'o' :as) | testHead (not . isId)  as -> Macro ~> next 5 as
  ('m':'a':'p':'r':'e':'d':'u':'c':'e':as) | testHead (not . isId)  as -> MapReduce ~> next 9 as
  ('i':'n'        :as) | testHead (not . isId)  as -> In     ~> next 2 as
  ('='            :as) | testHead (not . isSym) as -> Equals ~> next 1 as
  ('\\'           :as) | testHead (not . isSym) as -> Lambda ~> next 1 as
  ('@'            :as) | testHead (not . isSym) as -> At     ~> next 1 as
  (':'            :as) | testHead (not . isSym) as -> Colon  ~> next 1 as
  ('!'            :as) | testHead (not . isSym) as -> Bang   ~> next 1 as
  ('('            :as) -> LParen   ~> next 1 as
  (')'            :as) -> RParen   ~> next 1 as
  ('['            :as) -> LBracket ~> next 1 as
  (']'            :as) -> RBracket ~> next 1 as
  ('.'            :as) -> Dot      ~> next 1 as
  ('#'            :as) -> let (comment, rest) = span (/= '\n') as
                           in lexToks (n + length comment) rest
  as@(a:rest)
    | isLower a -> go Var isId as
    | isUpper a -> go Con isId as
    | isDigit a -> go (Num . read) isDigit as
    | isSym   a -> go Op isSym as
    | otherwise -> lexToks (n + 1) rest

  where
    tok ~> rest = (tok, n) : rest
    next k = lexToks (n + k)
    testHead :: (Char -> Bool) -> String -> Bool
    testHead _ ""    = True
    testHead f (a:_) = f a

    isId :: Char -> Bool
    isId c = isAlphaNum c || c == '_' || c == '\''

    isSym :: Char -> Bool
    isSym '(' = False
    isSym ')' = False
    isSym '[' = False
    isSym ']' = False
    isSym '#' = False
    isSym c   = isPunctuation c || isSymbol c

    go :: (String -> Tok) -> (Char -> Bool) -> String -> [(Tok, Int)]
    go c p xs = let (v, rest) = span p xs in (c v, n) : lexToks (n + length v) rest
