{-# LANGUAGE TemplateHaskell #-}
module ParseToAST where
import Control.Arrow
import Data.Map(Map)
import qualified Data.Map as M
import Data.List
import Data.Ord
import qualified Data.Set as S
import Data.Parser.Grempa.Static

import AST
import Constants
import Instr
import qualified Lexer
import Parser
import Type

splittings :: OpList a -> [(OpList a, Name, OpList a)]
splittings (Base _)             = []
splittings (Another e op rest) = (Base e, op, rest) :
  [(Another e op before, o, after) | (before, o, after) <- splittings rest]

fixities :: Show a => Map Name Fixity -> (a -> b) -> Map Name (b -> b -> b) -> OpList a -> b
fixities _ base _ (Base e) = base e
fixities f base ops oplist =
  let spls              = [(a, (op, look op f), b) | (a, op, b) <- splittings oplist]
      sorted@(lowest:_) = sortBy (comparing prec) spls
      same              = takeWhile ((== prec lowest) . prec) sorted
   in if all ((== assoc lowest) . assoc) same
      then let (before, (op, _), after) = case assoc lowest of
                 L -> last same
                 R -> head same
            in look op ops (fixities f base ops before) (fixities f base ops after)
      else error $ "fixities.same precedence, different assocativity: " ++ show oplist
  where
    prec  (_, (_, Infix _ p), _) = p
    prec  _                      = error "fixities.prec"
    assoc (_, (_, Infix a _), _) = a
    assoc  _                     = error "fixities.assoc"
    look op = M.findWithDefault (error $ "fixities.undefined: " ++ show op) op

exprToTerm :: Expr -> Term
exprToTerm = term
  where
    isInfix c = case cFixity c of
      Infix _ _ -> True
      _         -> False
    op :: Constant -> BaseTerm Term -> BaseTerm Term -> BaseTerm Term
    op c x y     = Term ((case cImpl c of
      Right t -> t
      Left _  -> Term $ Const c) :$ Term x) :$ Term y
    infixOps = M.filter isInfix constants
    termOpMap  = M.map op infixOps
    termFixMap = M.map cFixity infixOps
    typeOpMap  = M.fromList [(n, f) | (n, f, _) <- typeOperations]
    typeFixMap = M.fromList [(n, f) | (n, _, f) <- typeOperations]
    term pexpr = Term $ case pexpr of
      EApp e1 e2     -> term e1 :$ term e2
      EMacro f e1 e2 -> unTerm $ subst (M.singleton f $ term e1) $ term e2
      EMapReduce node num m red -> unTerm $ term $ mapReduce node num m red
      EOp oplist -> fixities termFixMap (unTerm . term) termOpMap oplist
      EVar v         -> case M.lookup v constants of
        Nothing -> Var v
        Just c  -> case cImpl c of
          Right t -> unTerm t
          Left _  -> Const c
      EFix (v, mt) e -> unTerm $ apps (Term $ Term Fix :$ abss ((v,fmap typeToType mt) : typfvs) t') tfvs
        where
          t = term e
          fvs = S.toList $ S.delete v $ fv t
          tfvs = map (Term . Var) fvs
          typfvs = map (\v' -> (v', Nothing)) fvs
          t' = subst (M.singleton v (apps (Term $ Var v) tfvs)) t
      EAtNode e n    -> term e :@ n
      EInt i         -> Lit $ lint i
      EFloat i1 i2   -> Lit $ lfloat i1 i2
      EArray e       -> Lit $ LArray (array $ term e)
      ELam vs e1     -> unTerm $ abss (map (second (fmap typeToType)) vs) (term e1)
      where
        abss = flip $ foldr (\(v, t) -> Term . Abs v t)
        apps = foldl (\t -> Term . (:$) t)
      --ECase e brs def -> Case (exprToTerm e) (map (fromIntegral *** exprToTerm) brs) $ exprToTerm def

    typeToType :: PType -> Type
    typeToType (PTCon c ts) = TConstr (typeConstr c) $ map typeToType ts
    typeToType (PTOp oplist) = fixities typeFixMap typeToType typeOpMap oplist

    typeConstr :: String -> TConstr
    typeConstr "Int"   = TInt
    typeConstr "Float" = TFloat
    typeConstr "Com"   = TCom
    typeConstr "[]"    = TArray
    typeConstr s       = error $ "Unknown type constructor: " ++ s


    -- Hackety hax
    array :: Term -> [PrimLit]
    array (Term (Term (_ :$ Term (Lit (PrimLit l))) :$ rest)) = l : array rest
    array (Term (Lit (PrimLit l)))                            = [l]
    array t = error $ "Array can only be initialised with literals: " ++ show t

mapReduce :: String -> Integer -> Expr -> Expr -> Expr
mapReduce nn num m red =
  foldr (\node e inner -> (EVar "new" `EApp` elam (varName node) (e inner)) `EAtNode` nodeName node)
    (\inner -> sequential
      (mapStep :
        (do nodes <- takeWhile ((>0) . length . fst)
                   $ zip (tail $ iterate reduceNodes nodeNames) (iterate (*2) 1)
            return $ reduceStep nodes
        )
        ++ [inner]
      )
    ) nodeNames (deref $ varName $ head nodeNames)
  where
    mapStep = inParallel $ do
      node <- nodeNames
      return $ ((m `EApp` EVar (varName node)) `EApp` EInt node) `EAtNode` nodeName node
    reduceStep (nodes, otherOffset) = inParallel $ do
      node <- nodes
      return $ (new "x" $ new "y" $ sequential
                [ assign "x" (deref $ varName node)
                , assign "y" (deref $ varName $ node + otherOffset)
                , ((red `EApp` EVar (varName node)) `EApp` deref "x") `EApp` deref "y"
                ]) `EAtNode` nodeName node
    varName node = "d_" ++ show node
    nodeName n = nn ++ show n
    nodeNames = [0..num - 1]
    reduceNodes (x:_:xs) = x : reduceNodes xs
    reduceNodes _        = []
    new x e    = EVar "new" `EApp` elam x e
    inParallel = foldr1 (\x y -> (EVar "||" `EApp` x) `EApp` y)
    sequential = foldr1 (\x y -> (EVar ";"  `EApp` x) `EApp` y)
    elam x = ELam [(x, Nothing)]
    deref = EApp (EVar "!") . EVar
    assign x y = (EVar ":=" `EApp` EVar x) `EApp` y

parseResult :: String -> Expr
parseResult str = case result of
  Left (ParseError expected pos) ->
    error $ unlines
      ["Unexpected token on line " ++ show (line + 1) ++ ", column " ++ show (col + 1) ++ "."
      ,theLine
      ,replicate (fromIntegral col) ' ' ++ "^"
      ,"Expected: "
      ,intercalate ", " (map (show . fmap Lexer.present) expected) ++ "."
      ]
    where
      realPos = (positions ++ repeat (length str)) !! fromIntegral pos
      lns = lines str
      (line, col) = lineRow realPos 0 lns
      theLine = lns !! line

  Left (InternalParserError pos) -> error $ "internal parser error at position "
                                        ++ show pos
  Right a -> a
  where
    (tokens, positions) = unzip $ Lexer.lexToks 0 str
    result = $(mkStaticParser fun [|fun|]) tokens
    lineRow _   _   [] = error "lineRow"
    lineRow pos num (l : ls)
      | len > pos = (num, pos)
      | otherwise = case ls of
        [] -> (num, pos)
        _  -> lineRow (pos - len) (num + 1) ls
      where len = length l + 1 -- + Line break

parseTerm :: String -> Term
parseTerm = exprToTerm . parseResult

