{-# LANGUAGE FlexibleInstances #-}
module Type where

import Control.Applicative
import Data.Foldable as F
import qualified Data.Map as M
import Data.Map(Map)
import Data.Monoid
import Data.Traversable

type TVar = String

infixr 5 :=>:
infixr 6 :*:

data TConstr = TInt | TFloat | TCom | TArray
  deriving (Eq, Ord, Show)

data Kind = KType | KArrow [Kind] Kind

btKind :: TConstr -> Kind
btKind bt = case bt of
  TInt -> KType
  TFloat -> KType
  TCom -> KType
  TArray -> [KType] `KArrow` KType

data TypeLike v c
  = TConstr c [TypeLike v c]
  | TypeLike v c :=>: TypeLike v c
  | TypeLike v c :*:  TypeLike v c
  | TVar v
  deriving (Eq, Ord, Show)

tbase :: c -> TypeLike v c
tbase = flip TConstr []

tint :: TypeLike v TConstr
tint = tbase TInt

tfloat :: TypeLike v TConstr
tfloat = tbase TFloat

tcom :: TypeLike v TConstr
tcom = tbase TCom

tvar :: TypeLike v TConstr -> TypeLike v TConstr
tvar a = a :*: (a :=>: tcom)

tmutarray :: TypeLike v TConstr -> TypeLike v TConstr
tmutarray a = tint :=>: tvar a

tarray :: TypeLike v TConstr -> TypeLike v TConstr
tarray t = TConstr TArray [t]

type Type = TypeLike TVar TConstr

newtype TypeOnVar c v = TypeOnVar {unTypeOnVar :: TypeLike v c}

instance Functor (TypeOnVar c) where
  fmap f (TypeOnVar typ) = TypeOnVar $ case typ of
    TConstr c ts -> TConstr c $ map rec ts
    t1 :=>: t2 -> rec t1 :=>: rec t2
    t1 :*:  t2 -> rec t1 :*:  rec t2
    TVar v     -> TVar $ f v
    where rec = unTypeOnVar . fmap f . TypeOnVar

instance Foldable (TypeOnVar c) where
  foldMap f (TypeOnVar typ) = case typ of
    TConstr _ ts -> mconcat $ map rec ts
    t1 :=>: t2 -> rec t1 `mappend` rec t2
    t1 :*:  t2 -> rec t1 `mappend` rec t2
    TVar v     -> f v
    where rec = foldMap f . TypeOnVar

instance Traversable (TypeOnVar c) where
  traverse f (TypeOnVar typ) = TypeOnVar <$> case typ of
    TConstr c ts -> TConstr c <$> sequenceA (map rec ts)
    t1 :=>: t2 -> (:=>:) <$> rec t1 <*> rec t2
    t1 :*:  t2 -> (:*:)  <$> rec t1 <*> rec t2
    TVar v     -> TVar <$> f v
    where rec = fmap unTypeOnVar . traverse f . TypeOnVar

newtype TypeOnConstr v c = TypeOnConstr {unTypeOnConstr :: TypeLike v c}

instance Functor (TypeOnConstr v) where
  fmap f (TypeOnConstr typ) = TypeOnConstr $ case typ of
    TConstr c ts -> TConstr (f c) $ map rec ts
    t1 :=>: t2 -> rec t1 :=>: rec t2
    t1 :*:  t2 -> rec t1 :*:  rec t2
    TVar v     -> TVar v
    where rec = unTypeOnConstr . fmap f . TypeOnConstr

instance Foldable (TypeOnConstr v) where
  foldMap f (TypeOnConstr typ) = case typ of
    TConstr c ts -> mconcat $ f c : map rec ts
    t1 :=>: t2 -> rec t1 `mappend` rec t2
    t1 :*:  t2 -> rec t1 `mappend` rec t2
    TVar _     -> mempty
    where rec = foldMap f . TypeOnConstr

instance Traversable (TypeOnConstr v) where
  traverse f (TypeOnConstr typ) = TypeOnConstr <$> case typ of
    TConstr c ts -> TConstr <$> f c <*> sequenceA (map rec ts)
    t1 :=>: t2 -> (:=>:) <$> rec t1 <*> rec t2
    t1 :*:  t2 -> (:*:)  <$> rec t1 <*> rec t2
    TVar v     -> pure $ TVar v
    where rec = fmap unTypeOnConstr . traverse f . TypeOnConstr

type TSubst = Map TVar Type

class TypeSubst a where
  typeSubst :: TSubst -> a -> a

instance (TypeSubst a, TypeSubst b) => TypeSubst (a, b) where
  typeSubst subst (a, b) = (typeSubst subst a, typeSubst subst b)

instance TypeSubst v => TypeSubst (Map k v) where
  typeSubst subst m = M.map (typeSubst subst) m

instance TypeSubst (TypeLike TVar TConstr) where
  typeSubst subst typ = case typ of
    TConstr c ts -> TConstr c $ map (typeSubst subst) ts
    t1 :=>: t2   -> typeSubst subst t1 :=>: typeSubst subst t2
    t1 :*:  t2   -> typeSubst subst t1 :*:  typeSubst subst t2
    TVar v       -> maybe typ (typeSubst subst) $ M.lookup v subst
