#ifndef __PRELUDE_H__
#define __PRELUDE_H__

#include <math.h>
#include "mpi.h"
#include <pthread.h>
#define __STDC_FORMAT_MACROS
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MALLOC malloc
#define FREE   free
#define REALLOC realloc

#ifdef NDEBUG
#define DEBUG_PRINTF(...) printf(__VA_ARGS__)
#else
#define DEBUG_PRINTF(...)
#endif
#define ERROR(...) fprintf(stderr, "********" __VA_ARGS__); exit(0);
/* #define PRINT_DEBUG_INFO printf("Entering %s\n", __FUNCTION__); print_registers_int_data(regs); printf("\n") */
#define PRINT_DEBUG_INFO

/*****************************************************************************/
/* An int of the architecture's pointer size */
typedef intptr_t int_t;

/* A float of the architecture's pointer size.
 * FIXME: This is non-portable but "probably" okay in most cases. */
#if __WORDSIZE == 64
typedef double floatptr_t;
#else
typedef float floatptr_t;
#endif

union Data_t;
struct HeapVal_t;
struct Mutex_t;
typedef union Data_t Data;
typedef struct HeapVal_t HeapVal;
typedef struct Mutex_t Mutex;


/*****************************************************************************
 * Init
 *****************************************************************************/
static inline void initialise(int argc, char** argv) {
  int provided;
  if(MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &provided) != MPI_SUCCESS) {
    ERROR("MPI_Init failed!\n");
  }
  if(provided != MPI_THREAD_MULTIPLE) {
    printf("******* The MPI environment could not provide the required level of\n"
           "******* thread support. This may lead to deadlocks.\n"
           "******* Proceeding...\n");
  }
}

/*****************************************************************************
 * Data
 *****************************************************************************/
union Data_t {
  HeapVal* ptr;
  int_t integer;
  floatptr_t floating;
  Mutex* mutex;
};

/*****************************************************************************
 * Registers
 *****************************************************************************/
#define MEMBER_SIZE(type, member) sizeof(((type*)0)->member)
#define REGISTERS_SIZE(n) (sizeof(Registers) + (n) * sizeof(Data))
#define REGISTERS_MESSAGE_SIZE(n) (MEMBER_SIZE(Registers, message) + (n) * sizeof(Data))
Data emptyReg;

struct Registers_t {
  Data temp;
  size_t dataLength;
  struct {
    Data ptr1;
    Data ptr2;
  } message;
  Data data[];
};

typedef struct Registers_t Registers;

static inline Registers* malloc_registers(size_t dataLength) {
  Registers* regs = MALLOC(REGISTERS_SIZE(dataLength));
  return regs;
}

static inline void initial_registers(Registers* regs, size_t dataLength) {
  memset(regs, 0, REGISTERS_SIZE(dataLength));
  regs->dataLength = dataLength;
}

static inline Registers* resize_registers(Registers* regs, size_t dataLength) {
  if(dataLength != regs->dataLength) {
    Registers* newRegs = REALLOC(regs, REGISTERS_SIZE(dataLength));
    if(newRegs) {
      newRegs->dataLength = dataLength;
      return newRegs;
    } else {
      ERROR("resize_registers failed\n");
      exit(0);
    }
  } else {
    return regs;
  }
}

static inline void print_registers_int_data(Registers* regs) {
  if(regs->dataLength == 0) {
    printf("(empty)");
    return;
  }
  printf("%" PRIdPTR, regs->data[0].integer);
  for(int i = 1; i < regs->dataLength; ++i) {
    printf(", %" PRIdPTR, regs->data[i].integer);
  }
}

static inline void print_registers_float_data(Registers* regs) {
  if(regs->dataLength == 0) {
    printf("(empty)");
    return;
  }
  printf("%f", regs->data[0].floating);
  for(int i = 1; i < regs->dataLength; ++i) {
    printf(", %f", regs->data[i].floating);
  }
}

/*****************************************************************************
 * Heap values
 *****************************************************************************/
#define HEAPVAL_SIZE(n) (sizeof(HeapVal) + sizeof(Data) * n)
struct HeapVal_t {
  Data fst;
  size_t dataLength;
  Data snd[];
};

static inline HeapVal* malloc_heapval(Data fst, int_t dataLength) {
  HeapVal* res = MALLOC(HEAPVAL_SIZE(dataLength));
  res->fst = fst;
  res->dataLength = dataLength;
  return res;
}

static inline HeapVal* new_heapval(Data fst, Data* snd, int_t dataLength) {
  HeapVal* res = malloc_heapval(fst, dataLength);
  res->snd[0] = snd[0];
  /*
  for(int i = 0; i < dataLength; ++i) {
    res->snd[i] = snd[i];
  }
  */
  return res;
}

static inline HeapVal* update_heapval(HeapVal* heapval, Data* newsnd, size_t dataLength) {
  if(dataLength != heapval->dataLength) {
    heapval = REALLOC(heapval, HEAPVAL_SIZE(dataLength));
    if(heapval) {
      heapval->dataLength = dataLength;
    } else {
      ERROR("update_heapval failed\n");
      exit(0);
    }
  }
  for(int i = 0; i < dataLength; ++i) {
    heapval->snd[i] = newsnd[i];
  }
  return heapval;
}

static inline void update_heap_array_index(HeapVal* source, int_t index, Data val) {
  source->snd[index] = val;
}

static inline Data access_heap_array_index(HeapVal* source, int_t index) {
  return source->snd[index];
}


static inline void get_heapval(HeapVal* heapval, Data* fst, Data* snd, size_t dataLength) {
  *fst = heapval->fst;
  for(int i = 0; i < dataLength; ++i) {
    snd[i] = heapval->snd[i];
  }
}

/*****************************************************************************
 * Mutex
 *****************************************************************************/
struct Mutex_t {
  pthread_mutex_t lock;
  int_t val;
};

static inline Mutex* malloc_mutex() {
  Mutex* mutex = MALLOC(sizeof(Mutex));

  int res = pthread_mutex_init(&mutex->lock, NULL);
  if(res) {
    ERROR("pthread_mutex_init failed with %d\n", res);
  }
  mutex->val = 0;
  return mutex;
}

static inline void free_mutex(Mutex* mutex) {
  int res = pthread_mutex_destroy(&mutex->lock);
  if(res) {
    ERROR("pthread_mutex_destroy failed with %d\n", res);
  }
  FREE(mutex);
}

static inline int_t atomic_incr(Mutex* mutex, int_t done) {
  pthread_mutex_lock(&mutex->lock);
  ++mutex->val;
  int_t val = mutex->val;
  int res = pthread_mutex_unlock(&mutex->lock);
  if(res) {
    ERROR("pthread_mutex_unlock failed with %d\n", res);
    exit(0);
  }
  if(val >= done) {
    free_mutex(mutex);
  }
  return val;
}

/*****************************************************************************
 * Thread spawning
 *****************************************************************************/
static inline Registers* code();

void* run_thread(void* vregs) {
  Registers* regs = (Registers*) vregs;
  regs = code(regs, regs->temp);
  FREE(regs);
  return NULL;
}

static inline void spawn_thread(Registers* regs, int label) {
  pthread_t pthread;
  regs->temp.integer = label;
  int res = pthread_create(&pthread, NULL, run_thread, regs);
  if(res) {
    ERROR("pthread_create failed with %d\n", res);
  }
}

static inline void fork_thread(Registers* regs, int label) {
  Registers* newRegs = malloc_registers(regs->dataLength);
  newRegs->dataLength = regs->dataLength;
  memcpy(&newRegs->message, &regs->message, REGISTERS_MESSAGE_SIZE(regs->dataLength));
  spawn_thread(newRegs, label);
}

/*****************************************************************************
 * Communication
 *****************************************************************************/
static inline void send(Registers* regs, int node, int port) {
  int myid;
  MPI_Comm_rank(MPI_COMM_WORLD, &myid);
#ifdef NDEBUG
  DEBUG_PRINTF("%d to %d: (%p, %p), ",
               myid, node,
               (void*) regs->message.ptr1.ptr,
               (void*) regs->message.ptr2.ptr);
  print_registers_int_data(regs);
  DEBUG_PRINTF("\n");
#endif
  MPI_Send(&regs->message,
           REGISTERS_MESSAGE_SIZE(regs->dataLength) / sizeof(int32_t),
           MPI_INT, node, port, MPI_COMM_WORLD);
}

static inline int receive(Registers** regs) {
  MPI_Status status;
  MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
  int count;
  MPI_Get_count(&status, MPI_INT, &count);
  if(count == MPI_UNDEFINED) {
    ERROR("Receive count is undefined!\n");
  }
  size_t dataLength = (count * sizeof(int32_t) - REGISTERS_MESSAGE_SIZE(0)) / sizeof(Data);
  *regs = malloc_registers(dataLength);
  (*regs)->dataLength = dataLength;
  MPI_Recv(&(*regs)->message, count, MPI_INT, MPI_ANY_SOURCE,
           MPI_ANY_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
  int myid;
  MPI_Comm_rank(MPI_COMM_WORLD, &myid);
  DEBUG_PRINTF("%d rec on port %d\n", myid, status.MPI_TAG);
  return status.MPI_TAG;
}

/*****************************************************************************
 * Primops
 *****************************************************************************/
#define BINOP(op, length, field) \
  for(int i = 0; i < (length); ++i) { \
    regs->data[i].field = regs->message.ptr1.ptr->snd[i].field op regs->data[i].field; \
  }
#define BINOP_ALL(op, length, field) \
  regs->data[0].integer = regs->message.ptr1.ptr->snd[0].field op regs->data[0].field; \
  for(int i = 1; i < (length); ++i) { \
    regs->data[0].integer = regs->data[0].integer && (regs->message.ptr1.ptr->snd[i].field op regs->data[i].field); \
  } \
  resize_registers(regs, 1);

#define BINOP_ANY(op, length, field) \
  regs->data[0].integer = regs->message.ptr1.ptr->snd[0].field op regs->data[0].field; \
  for(int i = 1; i < (length); ++i) { \
    regs->data[0].integer = regs->data[0].integer || (regs->message.ptr1.ptr->snd[i].field op regs->data[i].field); \
  } \
  resize_registers(regs, 1);

void print_int(int_t i) {
  printf("%" PRIdPTR "\n", i);
}
void print_float(floatptr_t f) {
  printf("%f\n", f);
}

#endif
