{-# LANGUAGE TypeOperators #-}
module AST where

import Control.Applicative
import Control.Arrow
import Data.Map(Map)
import qualified Data.Map as M
import Data.Maybe
import Data.Monoid

import Component(Node)

infixr 5 :->:
infix  2 :::
infixl 4 :$
infixl 3 :@

type Var = String

data Type
  = TInt
  | Type :->: Type
  | TVar Var
  deriving (Eq, Ord, Show)

type TSubst = Map Var Type

class TypeSubst a where
  typeSubst :: TSubst -> a -> a

instance TypeSubst Type where
  typeSubst subst typ = case typ of
    TInt       -> TInt
    t1 :->: t2 -> typeSubst subst t1 :->: typeSubst subst t2
    TVar v     -> maybe typ (typeSubst subst) $ M.lookup v subst

occurs :: Var -> Type -> Bool
occurs v typ = case typ of
  TInt       -> False
  t1 :->: t2 -> occurs v t1 || occurs v t2
  TVar v'    -> v == v'

unify :: Type -> Type -> TSubst
unify t1 t2 = case (t1, t2) of
  (TInt, TInt)                 -> mempty
  (t1 :->: t2, s1 :->: s2)     -> u2 `mappend` u1
    where u1 = unify t1 s1
          u2 = unify (typeSubst u1 t2) (typeSubst u1 s2)
  (TVar v, TVar v') | v == v' -> mempty
  (TVar v, t)                 -> varCase v t
  (t, TVar v)                 -> varCase v t
  _                           -> error $ "can't unify " ++ show (t1, t2)
  where
    varCase v t = if occurs v t
      then error $ "unification occurs check: " ++ v ++ " occurs in " ++ show t
      else M.singleton v t

data BaseTerm term
  = Var Var
  | Abs Var (Maybe Type) term
  | term :$ term
  | Lit Int
  | Plus | Times | Minus
  | Fix
  -- | Case term [(Int, term)] term
  | If
  | term :@ Node
  deriving (Eq, Ord, Show)

instance TypeSubst term => TypeSubst (BaseTerm term) where
  typeSubst subst term = case term of
    Var v      -> Var v
    Abs v mt t -> Abs v (rec <$> mt) $ rec t
    t1 :$ t2   -> rec t1 :$ rec t2
    Lit n      -> Lit n
    Plus       -> Plus
    Times      -> Times
    Minus      -> Minus
    t :@ n     -> rec t :@ n
    Fix        -> Fix
    If         -> If
    -- Case t b d -> Case (rec t) (map (second rec) b) (rec d)
    where
      rec :: TypeSubst a => a -> a
      rec = typeSubst subst

data a ::: b = a ::: b
  deriving (Eq, Ord, Show)
instance (TypeSubst a, TypeSubst b) => TypeSubst (a ::: b) where
  typeSubst subst (a ::: b) = typeSubst subst a ::: typeSubst subst b
instance (TypeSubst a, TypeSubst b) => TypeSubst (a, b) where
  typeSubst subst (a, b) = (typeSubst subst a, typeSubst subst b)

newtype Term = Term (BaseTerm Term)
  deriving (Eq, Ord, Show)
unTerm :: Term -> BaseTerm Term
unTerm (Term t) = t
instance TypeSubst Term where
  typeSubst subst (Term t) = Term $ typeSubst subst t

newtype TypedTerm = TypedTerm (BaseTerm (TypedTerm ::: Type))
  deriving (Eq, Ord, Show)
unTypedTerm :: TypedTerm -> BaseTerm (TypedTerm ::: Type)
unTypedTerm (TypedTerm t) = t
instance TypeSubst TypedTerm where
  typeSubst subst (TypedTerm t) = TypedTerm $ typeSubst subst t

instance TypeSubst v => TypeSubst (Map k v) where
  typeSubst subst m = M.map (typeSubst subst) m
