{-# LANGUAGE TypeOperators #-}
module ComponentCompiler where
import Control.Applicative
import Control.Monad
import qualified Data.Map as M
import Data.Map(Map)
import Data.Maybe
import Data.Monoid
import qualified Data.Set as S
import Data.Set(Set)
import Language.C.Generate as C hiding ((:->), (:>))

import Component
import Compiler
import CPrelude as CP

-- | Get a Map where each element maps to an element that represents its equivalence
--   class given a list of equalities
eqClasses :: Ord a => [(a, a)] -> a :-> a
eqClasses = M.map (head . S.toList) . foldr equate mempty
  where
    equate :: Ord a => (a, a) -> Map a (Set a) -> Map a (Set a)
    equate (p1, p2) m = M.fromList [(p, newSet) | p <- S.toList newSet] `M.union` m
      where
        ps     = S.fromList [p1, p2]
        newSet = fromMaybe ps (M.lookup p1 m) `S.union`
                 fromMaybe ps (M.lookup p2 m)

prop_eqClassesStable :: [(Int, Int)] -> Bool
prop_eqClassesStable xs = eqClasses xs Prelude.== eqClasses (M.toList $ eqClasses xs)

connectPorts :: [(Port, Port)] -> [ComponentCode] -> [ComponentCode]
connectPorts ps ccs = map (substPortsCC $ eqClasses ps) ccs

defineLabels :: ComponentCode -> Decl (Label :-> Fun (IO ()))
defineLabels (Listen ps ls) = do
  M.fromList <$> forM (M.keys ls) (\l -> do
    f <- declareFun (l ++ "x")
    return (l, f))

portComponents :: [(ComponentCode, Node)] -> Port :-> [Node]
portComponents = M.unionsWith (++) . map (uncurry component)
  where
    component :: ComponentCode -> Node -> Port :-> [Node]
    component (Listen ps ls) nodeid = M.map (const [nodeid]) ps

ccompile :: (ComponentCode, Node)
         -> Label :-> Fun (IO ())
         -> Port  :-> [Node]
         -> Node  :-> Int
         -> Decl ()
ccompile (Listen ps ls, nodeid) lfs portNodes nodeInts = do
  mapM_ (uncurry defineLabel) (M.toList ls)
  makeMain $ \_ argc argv -> do
    scall initialise argc argv
    case (M.toList ps, M.toList ls) of -- Special hack!
      ([(p, l1)], [(l2, Print)]) -> compileInstrs (Send p)
      _ -> return ()
    while (bool True) $ do
      port <- "port" =. call receive
      switch port
        [(p, do scall $ lfs !! label; C.break) | (p, label) <- M.toList ps]
        C.break
    ret $ lit 0
  return ()
  where
    defineLabel :: Label -> Code -> Decl ()
    defineLabel l c = do
      let func = lfs !! l
      defineFun func () $ do
        stmt $ trustMe "PRINT"
        compileInstrs c :: Stmt () ()

    compileInstrs :: Code -> Stmt r ()
    compileInstrs is = case is of
      instr :> rest -> do
        compileInstr instr
        compileInstrs rest
      Send p -> scall send (lit $ portnode p) (lit p)
      Jump l -> scall (lfs !! l)
      Match l1 l2 -> scall match (funPtr $ lfs !! l1)
                                 (funPtr $ lfs !! l2)
      Done        -> scall CP.done
      Print       -> scall CP.print
      Switch brs def -> switch (trustMe "data.data.integer") [(n, do scall (lfs !! l); C.break) | (n, l) <- brs] (do stmt (trustMe "data = Empty()"); scall (lfs !! def); C.break)

    compileInstr :: Instr -> Stmt r ()
    compileInstr instr = case instr of
      Inl   -> scall CP.inl
      Inr   -> scall CP.inr
      Fst   -> scall CP.fst
      Snd   -> scall CP.snd
      UnFst -> scall CP.unfst
      UnSnd -> scall CP.unsnd
      Flip  -> scall CP.flip
      Push  -> scall CP.push
      Pop   -> scall CP.pop
      Int n -> scall CP.integer (lit n)
      Add   -> scall CP.add
      Mul   -> scall CP.mul
      Sub   -> scall CP.sub
      PushInt -> scall CP.pushint


    err :: String -> a
    err s  = error $ "ccompile: " ++ s
    (!!) :: (Show k, Ord k) => Map k v -> k -> v
    m !! l = fromMaybe (err $ show l) $ M.lookup l m
    portnode :: Port -> Int
    portnode p = case M.lookup p portNodes of
      Just xs -> case filter (Prelude./= nodeid) xs of
        [n] | M.member n nodeInts ->  nodeInts !! n
        x   -> err $ nodeid ++ " neq " ++ show x
      Nothing -> err $ show p

compileComponents :: [(ComponentCode, Node)] -> [(Node, Decl ())]
compileComponents ccs =
  let pcs      = portComponents ccs
      nodeInts = M.fromList $ zip (map Prelude.snd ccs) [0..]
   in
    [(node, do includePrelude
               lfs <- defineLabels cc
               ccompile (cc, node) lfs pcs nodeInts)
    | (cc, node) <- ccs]
